<?php

declare(strict_types=1);

namespace Ifacesoft\Ice\Query\Application\UseCase;

use Ifacesoft\Ice\Data\Interfaces\ValueObject\Scalar\StringValue\XmlStringInterface;
use Ifacesoft\Ice\Generator\Domain\Interfaces\GeneratorInterface;
use Ifacesoft\Ice\Query\Domain\Collection\Dto\Table\ColumnCollection;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\ClCollection\Cl;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\ClCollection\Cl\DTCollection\DT;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\FkCollection\Fk;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\FkCollection\Fk\ClPrCollection\ClPr;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\FkCollection\Fk\PrTbCollection\PrTb;
use Ifacesoft\Ice\Query\Domain\Dto\Dbwrench\TblCollection\Tbl\PkCollection\Pk;
use Ifacesoft\Ice\Query\Domain\Dto\Table\Column;
use Ifacesoft\Ice\Query\Domain\Model\Class\Php8\Doctrine\Entity;
use Ifacesoft\Ice\Query\Domain\ValueObject\Table;
use Ifacesoft\Ice\Render\Domain\Interfaces\TemplaterInterface;
use JsonException;
use ReflectionException;
use SimpleXMLElement;

class GenerateEntitiesFromDbwrenchUseCase
{
    private GeneratorInterface $generator;
    private TemplaterInterface $templater;

    public function __construct(
        GeneratorInterface $generator,
        TemplaterInterface $templater
    ) {
        $this->generator = $generator;
        $this->templater = $templater;
    }

    /**
     * @throws JsonException
     * @throws ReflectionException
     */
    public function __invoke(XmlStringInterface $xmlString): void
    {
        /** @var array<string, mixed> $properties */
        $properties = $this->xmlToArray($xmlString->toSimpleXmlElement()->Sch); // @phpstan-ignore property.nonObject

        $dbwrench = Dbwrench::newInstance($properties);

        /** @var Dbwrench\TblCollection\Tbl $tbl */
        foreach ($dbwrench->getTbl() as $tbl) {
            $tableName = new Table($tbl->getAttributes()->getNm());

            if (!$tableName->getPrefix()) {
                continue;
            }

            /** @var Pk $pk */
            $pk = $tbl->getPk()->get(0);
            $pkColumnName = $pk->getAttributes()->getClNs();

            $joinColumns = [];

            if ($fkCollection = $tbl->getFk()) {
                /** @var Fk $fk */
                foreach ($fkCollection as $fk) {
                    /** @var ClPr $clPr */
                    $clPr = $fk->getClPr()->get(0);

                    /** @var PrTb $prTb */
                    $prTb = $fk->getPrTb()->get(0);

                    $joinColumns[$clPr->getAttributes()->getCdCl()] = new Column\Join(
                        new Table($prTb->getAttributes()->getNm()),
                        $clPr->getAttributes()->getPrCl()
                    );
                }
            }

            $columnCollection = new ColumnCollection();

            /** @var Cl $cl */
            foreach ($tbl->getCl() as $cl) {
                $columnMain = $cl->getAttributes();
                $defaultValue = $columnMain->getDf();

                /** @var DT $dt */
                $dt = $cl->getDT()->get(0);
                $columnProps = $dt->getAttributes();

                $columnCollection = $columnCollection->add(
                    new Column(
                        $columnMain->getNm(),
                        $columnProps->getDs(),
                        (bool) $columnMain->getNu(),
                        'null' === $columnProps->getLn() ? 0 : (int) $columnProps->getLn(),
                        $defaultValue ?: null,
                        $columnMain->getNm() === $pkColumnName,
                        $joinColumns[$columnMain->getNm()] ?? null
                    )
                );
            }

            (
            new Entity(
                $tableName,
                $columnCollection,
                'Entity\\'
            )
            )->generate($this->generator, $this->templater);
        }
    }

    /**
     * @return array<mixed>|string
     */
    private function xmlToArray(SimpleXMLElement $xmlObject)
    {
        $array = [];

        // Преобразуем атрибуты элемента в ассоциативный массив
        if ($xmlObject->attributes()) {
            foreach ($xmlObject->attributes() as $key => $value) {
                $array['attributes'][$key] = (string) $value;
            }
        }

        // Преобразуем дочерние элементы в массивы
        if ($xmlObject->children()) {
            foreach ($xmlObject->children() as $key => $child) {
                $array[$key][] = $this->xmlToArray($child);
            }

            foreach ($array as &$value) {
                if (is_array($value) && 1 === count($value) && !is_array(reset($value))) {
                    $value = reset($value);
                }
            }
            unset($value);
        } else {
            $array = (string) $xmlObject;
        }

        return $array;
    }
}
