<?php

declare(strict_types=1);

namespace Ifacesoft\Ice\Query\Domain\Model\Class\Php8\Doctrine;

use Ifacesoft\Ice\Core\Domain\Model\Module;
use Ifacesoft\Ice\Generator\Domain\Abstracts\AbstractClass;
use Ifacesoft\Ice\Generator\Domain\Interfaces\GeneratorInterface;
use Ifacesoft\Ice\Query\Domain\Collection\Dto\Table\ColumnCollection;
use Ifacesoft\Ice\Query\Domain\Dto\Table\Column;
use Ifacesoft\Ice\Query\Domain\ValueObject\Table;
use Ifacesoft\Ice\Render\Domain\Interfaces\TemplaterInterface;
use ReflectionException;

final class Entity extends AbstractClass
{
    public const SOURCES_PATH = 'src/';
    public const TEMPLATES_PATH = 'templates/';
    public const CALSS_FILE_EXT = '.php';
    public const TEMPLATE_FILE_EXT = '.tpl';
    public const ENTITY_NAMESPACE = 'Domain\\';
    public const FACTORY_INTERFACE_NAMESPACE = 'Interfaces\Factory\\';
    public const FACTORY_INTERFACE_POSTFIX = 'FactoryInterface';
    public const FACTORY_NAMESPACE = 'Infrastructure\Factory\\';
    public const FACTORY_POSTFIX = 'Factory';
    public const MANAGER_INTERFACE_NAMESPACE = 'Interfaces\Manager\\';
    public const MANAGER_INTERFACE_POSTFIX = 'ManagerInterface';
    public const MANAGER_NAMESPACE = 'Infrastructure\Manager\\';
    public const MANAGER_POSTFIX = 'Manager';
    public const REPOSITORY_INTERFACE_NAMESPACE = 'Interfaces\Repository\\';
    public const REPOSITORY_INTERFACE_POSTFIX = 'RepositoryInterface';
    public const REPOSITORY_NAMESPACE = 'Infrastructure\Repository\\';
    public const REPOSITORY_POSTFIX = 'Repository';

    private Table $tableName;

    private string $classNamespace;

    /**
     * @param ColumnCollection<int, Column> $columns
     *
     * @throws ReflectionException
     */
    public function __construct(
        Table            $tableName,
        ColumnCollection $columns,
        string           $classNamespace,
    ) {
        $this->tableName = $tableName;
        $this->classNamespace = $classNamespace;

        parent::__construct($tableName->getType(self::ENTITY_NAMESPACE . $classNamespace), $columns);
    }

    /**
     * @throws ReflectionException
     */
    public function generate(GeneratorInterface $generator, TemplaterInterface $templater): self
    {
        $class = $this->getType();

        /** @var Module $module */
        $module = $class->getIceModule();

        $params = [
            'namespace' => $class->getNamespace(),
            'moduleNamespace' => $module->getNamespace(),
            'class' => $class->getValue(),
            'className' => $class->getClassName(),
            'tableName' => $this->tableName->getValue(),
            'constructorArguments' => $this->getConstructorArgumentsCode($this->getColumns()),
        ];

        $this->generateEntity($generator, $templater, $params);
        $this->generateEntityFactoryInterface($generator, $templater, $params);
        $this->generateEntityManagerInterface($generator, $templater, $params);
        $this->generateEntityRepositoryInterface($generator, $templater, $params);
        $this->generateEntityFactory($generator, $templater, $params);
        $this->generateEntityManager($generator, $templater, $params);
        $this->generateEntityRepository($generator, $templater, $params);

        //                            $this->tableName->getType(
        //                                self::MANAGER_NAMESPACE . $this->classNamespace,
        //                                self::MANAGER_POSTFIX
        //                            )->getValue(),
        //                            $this->tableName->getType(
        //                                self::REPOSITORY_NAMESPACE . $this->classNamespace,
        //                                self::REPOSITORY_POSTFIX
        //                            )->getValue(),
        //                        ]);

        return $this;
    }

    /**
     * @return ColumnCollection<int, Column>
     */
    public function getColumns(): ColumnCollection
    {
        return $this->getProperties(); // @phpstan-ignore return.type
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    public function generateEntity(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $generator->generate(
            $this->getType()->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $this->getIceType()->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            ),
            true
        );
    }

    /**
     * @param ColumnCollection<int, Column> $columns
     *
     * @throws ReflectionException
     */
    private function getConstructorArgumentsCode(ColumnCollection $columns): string
    {
        $code = '';

        /** @var Column $column */
        foreach ($columns as $column) {
            $doctrineType = ', type: \'' . $column->getDataType()->getDoctrineType() . '\'';
            $length = $column->getDataType()->getLength() ? ', length: ' . $column->getDataType()->getLength() : '';
            $nullable = ', nullable: ' . ($column->getDataType()->isNullable() ? 'true' : 'false');

            if ($column->isPk()) {
                $code .= str_repeat(' ', 8) . '#[ORM\Id]' . "\n";
            }

            if ($fk = $column->getFk()) {
                $joinClassName = $fk->getTableName()->getClassName();

                $code .= str_repeat(' ', 8) . '#[ORM\ManyToOne(targetEntity: ' . $fk->getTableName()->getClassName() . '::class)]' . "\n";
                $code .= str_repeat(' ', 8) . '#[ORM\JoinColumn(name: \'' . $column->getName() . '\', referencedColumnName: \'' . $fk->getColumnName() . '\'' . $nullable . ')]' . "\n";
                $code .= str_repeat(' ', 8) . 'public readonly ' . ($column->getDataType()->isNullable() ? '?' : '') . $joinClassName . ' $' . $column->getAlias() . ",\n\n";
            } else {
                $code .= str_repeat(' ', 8) . '#[ORM\Column(name: \'' . $column->getName() . '\'' . $doctrineType . $length . $nullable . ')]' . "\n";
                $code .= str_repeat(' ', 8) . 'public readonly ' . ($column->getDataType() ? '?' : '') . $column->getDataType()->getPhpType()->getValue() . ' $' . $column->getAlias() . ",\n\n";
            }
        }

        return substr($code, 0, -2);
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityFactoryInterface(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $interfaceType = $this->tableName->getType(
            self::ENTITY_NAMESPACE . self::FACTORY_INTERFACE_NAMESPACE . $this->classNamespace,
            self::FACTORY_INTERFACE_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::FACTORY_INTERFACE_POSTFIX);

        $params['namespace'] = $interfaceType->getNamespace();

        $generator->generate(
            $interfaceType->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityManagerInterface(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $interfaceType = $this->tableName->getType(
            self::ENTITY_NAMESPACE . self::MANAGER_INTERFACE_NAMESPACE . $this->classNamespace,
            self::MANAGER_INTERFACE_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::MANAGER_INTERFACE_POSTFIX);

        $params['namespace'] = $interfaceType->getNamespace();

        $generator->generate(
            $interfaceType->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityRepositoryInterface(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $interfaceType = $this->tableName->getType(
            self::ENTITY_NAMESPACE . self::REPOSITORY_INTERFACE_NAMESPACE . $this->classNamespace,
            self::REPOSITORY_INTERFACE_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::REPOSITORY_INTERFACE_POSTFIX);

        $params['namespace'] = $interfaceType->getNamespace();

        $generator->generate(
            $interfaceType->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityFactory(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $type = $this->tableName->getType(
            self::FACTORY_NAMESPACE . $this->classNamespace,
            self::FACTORY_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::FACTORY_POSTFIX);

        $params['namespace'] = $type->getNamespace();

        $generator->generate(
            $type->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityManager(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $type = $this->tableName->getType(
            self::MANAGER_NAMESPACE . $this->classNamespace,
            self::MANAGER_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::MANAGER_POSTFIX);

        $params['namespace'] = $type->getNamespace();

        $generator->generate(
            $type->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }

    /**
     * @param array<string, string> $params
     *
     * @throws ReflectionException
     */
    private function generateEntityRepository(GeneratorInterface $generator, TemplaterInterface $templater, array $params): void
    {
        $type = $this->tableName->getType(
            self::REPOSITORY_NAMESPACE . $this->classNamespace,
            self::REPOSITORY_POSTFIX
        );

        $interfaceTemplateType = $this->getIceType(self::REPOSITORY_POSTFIX);

        $params['namespace'] = $type->getNamespace();

        $generator->generate(
            $type->getFile(self::SOURCES_PATH, self::CALSS_FILE_EXT),
            $templater->render(
                $interfaceTemplateType->getFile(self::TEMPLATES_PATH, self::TEMPLATE_FILE_EXT, $this->getIceModule()),
                $params
            )
        );
    }
}
